// TQKEY v1.0
//
// Program emulates the custom keyboard used by the laser game Thayer's Quest,
// using a standard PC/keyboard via the PC parallel port.
//
// by Warren Ondras, 4/28/01
//
// connect the TQ keyboard port to the PC parallel port as follows:

// TQ Pin:     Parallel port pin:
// R (+5V)		NOT connected
// Y (clock)	10 (_ACK)
// G (data)    1  (_STROBE)
// B (ground)  25 (GND)

#include <stdio.h>
#include <dos.h>

#define PARPORT 0x378	//constant for now, maybe configurable in the future?
#define TIMEOUT 10000   //how long we should wait for clock pulse from TQ hardware

//CONSTANTS:

//TQKeys maps each key in the TQ Matrix to a PC keyboard scancode
//(perhaps use a config file in the future)
const int TQKeys[40]={ 60, 59, 16,  2, 44, 30, 17,  3, 45, 31, 18,  4, 46, 32, 19,  5, 47, 33, 20,  6, 48, 34, 21,  7, 49, 35, 22,  8, 50, 36, 23,  9, 62, 37, 24, 10, 61, 38, 25, 11};
//  PC key:            F2  F1   Q   1,  Z   A   W   2   X   S   E   3   C   D   R   4   V   F   T   5   B   G   Y   6   N   H   U   7   M   J   I   8  F4   K   O   9  F3   L   P   0
//  TQ matrix:         F2  F1   Q   1,  Z   A   W   2   X   S   E   3   C   D   R   4   V   F   T   5   B   G   Y   6   N   H   U   7   M   J   I   8  F4   K   O   9  F3   L   P   0
//  (keynames are identical for now)


//GLOBALS:
bool KeyArray[128];  //stores state of all keyboard keys
int ScanCode=0;  //tmp value used by kbd interrupt handler

bool TQError;         // flag indicates a problem accessing TQ hardware
bool QuitFlag=false;  // flag set by keyboard handler when user presses ESC
void (_interrupt _far *old_kbd_handler)();  //stores original keyboard IRQ handler


//PROTOTYPES:
void _interrupt _far new_kbd_handler(); //our custom kbd handler
void SendBit(bool bit);


//MAIN:
int main(int argc, char *argv[])
{
int i;
int j;
int arraypos=0;

//startup message
printf("TQKey v1.0   4/28/01\r\n");
printf("By Warren Ondras (warren@dragons-lair-project.com)\r\n");
printf("Special thanks to Mark Broadhead (mark_broadhead@usa.net)\r\n");
printf("Updates and info at http://www.dragons-lair-project.com\r\n");
printf("\r\nPress ESC to exit\r\n\r\n");

//init keyarray
for (i=0;i<128;i++)
  KeyArray[i]=false;

// install kbd handler
old_kbd_handler = _dos_getvect(0x09);  //remember the existing keyboard handler
_dos_setvect(0x09, new_kbd_handler);


//main loop
while(!QuitFlag) {  // exit if ESC key was pressed

/* TQ Keyboard interface format:
  (thanks MarkB!)
{{1, 1, 0, 1, 0, F2, F1, Q, 1, 0},
{1, 1, 0, 1, 0,  Z,  A, W, 2, 0},
{1, 1, 0, 1, 0,  X,  S, E, 3, 0},
{1, 1, 0, 1, 0,  C,  D, R, 4, 0},
{1, 1, 0, 1, 0,  V,  F, T, 5, 0},
{1, 1, 0, 1, 0,  B,  G, Y, 6, 0},
{1, 1, 0, 1, 0,  N,  H, U, 7, 0},
{1, 1, 0, 1, 0,  M,  J, I, 8, 0},
{1, 1, 0, 1, 0, F4,  K, O, 9, 0},
{1, 1, 0, 1, 1, F3,  L, P, 0, 0}};   */

//send matrix info to TQ

//send the rows of the matrix, plus start/stop bits

arraypos=0; //array position counter
for (i=0;i<10;i++)  {
	SendBit(true);
	SendBit(true);
	SendBit(false);
	SendBit(true);
	SendBit(i==9);  //last row has extra bit set high
	for (j=0;j<4;j++)  {
		SendBit(KeyArray[TQKeys[arraypos]]);
	   arraypos++;
		}
   SendBit(false);
	}

if (TQError) {
	printf("Error communicating with TQ board\r\n");
   TQError=false;
	}
}

//remove kbd handler
_dos_setvect(0x09, old_kbd_handler);


} //end main


//custom keyboard interrupt handler
//stores state of all keys in global KeyArray
void _interrupt _far new_kbd_handler() {


//high-speed assembly to get key event from hardware
//NOTE:  based on code from "Black Art Of 3D Game Programming," Waite Group Press, 1995)
_asm   {
	sti                    // re-enable interrupts
	in al, 0x60      // get the key that was pressed from the keyboard
	xor ah,ah              // zero out upper 8 bits of AX
	mov ScanCode, ax  // store the key in global variable
	in al, 0x61     // set the control register to reflect key was read
	or al, 82h             // set the proper bits to reset the keyboard flip flop
	out 0x61,al     // send the new data back to the control register
	and al,7fh
	out 0x61,al     // complete the reset
	mov al,20h             // reset command
	out 0x20,al        // tell PIC to re-enable interrupts
	}

// update the keyboard table
KeyArray[(ScanCode & 0x7F)] = !(ScanCode & 0x80);
// scancode is returned with high bit 0 if it's a 'make' code (key press), or
// 1 if it's a 'break' code (key release)
// so we strip the high bit to get the array position, and
// store 'true' if high bit is 0, 'false' if 1

if (ScanCode==1) QuitFlag=true; //tell main loop to exit if ESC was pressed.

} // end Keyboard_Driver



// waits for clock pulse, sends a bit
void SendBit(bool bit) {
int i;
bool errflag;

//wait for clock line to be low
i=0;
errflag=true; //clear this if things work ok
while (i<TIMEOUT) {
   i++;
	if (inp(PARPORT+1)&0x40) {
   	i=TIMEOUT;  //end the timeout loop
      errflag=false;
      }
	}

if (!errflag) {
	//when clock line goes high, send the bit!
	i=0;
	errflag=true; //clear this if things work ok
	while (i<TIMEOUT) {
	   i++ ;
		if (!(inp(PARPORT+1)&0x40)) {
   			//set the parallel port bit
				outp(PARPORT+2, 0x01 & !(bit)); //bit 0 = _STROBE (inverted), the others shouldn't matter
	   	i=TIMEOUT; //end the timeout loop
	      errflag=false;
	      }
		}
	}

if (errflag) TQError=true;  //something didn't work, so set the global flag

}


